import time
from uuid import uuid4
from collections import defaultdict
from typing import List, Dict, TypeVar, Optional, Type, Generic, Union, Any
from ..models import T
from ..common import (
    QueryParser, QueryDomain, EntityNotFoundError)
from .repository import Repository


class MemoryRepository(Repository, Generic[T]):
    def __init__(self,  parser: QueryParser) -> None:
        self.data: Dict[str, Dict[str, T]] = defaultdict(dict)
        self.parser: QueryParser = parser
        self.max_items = 10_000

    def add(self, item: Union[T, List[T]]) -> List[T]:
        items = item if isinstance(item, list) else [item]
        for item in items:
            item.id = item.id or str(uuid4())
            item.updated_at = int(time.time())
            existing_item = self.data[self._location].get(item.id)
            if existing_item:
                item.created_at = existing_item.created_at
            else:
                item.created_at = item.updated_at

            self.data[self._location][item.id] = item
        return items

    def search(self, domain: QueryDomain,
               limit=10_000, offset=0) -> List[T]:
        items = []
        filter_function = self.parser.parse(domain)
        for item in list(self.data[self._location].values()):
            if filter_function(item):
                items.append(item)

        if offset is not None:
            items = items[offset:]

        if limit is not None:
            items = items[:min(limit, self.max_items)]

        return items

    def remove(self, item: Union[T, List[T]]) -> bool:
        items = item if isinstance(item, list) else [item]
        deleted = False
        for item in items:
            deleted_item = self.data[self._location].pop(item.id, None)
            deleted = bool(deleted_item) or deleted

        return deleted

    def count(self, domain: QueryDomain = None) -> int:
        count = 0
        domain = domain or []
        filter_function = self.parser.parse(domain)
        for item in list(self.data[self._location].values()):
            if filter_function(item):
                count += 1
        return count

    def load(self, data: Dict[str, Dict[str, T]]) -> None:
        self.data = data

    @property
    def _location(self) -> str:
        return 'data'



from estimark.application.domain.models import (
    Classification, Classifier, Link, Schedule, Slot, Task)
from estimark.application.domain.common import QueryParser
from estimark.application.domain.repositories import (
    ClassificationRepository,  ClassifierRepository, LinkRepository,
    ScheduleRepository, SlotRepository, TaskRepository)
from estimark.core.data.json import JsonRepository


class JsonClassificationRepository(
        JsonRepository[Classification], ClassificationRepository):
    """Json Classification Repository"""

    def __init__(self, file_path: str, parser: QueryParser,
                 collection_name: str='classifications',
                 file_suffix: str='') -> None:
        super().__init__(file_path, parser, collection_name,
                         Classification, file_suffix)


class JsonClassifierRepository(
        JsonRepository[Classifier], ClassifierRepository):
    """Json Classifier Repository"""

    def __init__(self, file_path: str, parser: QueryParser,
                 collection_name: str='classifiers',
                 file_suffix: str='') -> None:
        super().__init__(file_path, parser, collection_name,
                         Classifier, file_suffix)


class JsonLinkRepository(
        JsonRepository[Link], LinkRepository):
    """Json Link Repository"""

    def __init__(self, file_path: str, parser: QueryParser,
                 collection_name: str='links',
                 file_suffix: str='') -> None:
        super().__init__(file_path, parser, collection_name,
                         Link, file_suffix)


class JsonScheduleRepository(
        JsonRepository[Schedule], ScheduleRepository):
    """Json Schedule Repository"""

    def __init__(self, file_path: str, parser: QueryParser,
                 collection_name: str='schedules',
                 file_suffix: str='') -> None:
        super().__init__(file_path, parser, collection_name,
                         Schedule, file_suffix)


class JsonSlotRepository(
        JsonRepository[Slot], SlotRepository):
    """Json Slot Repository"""

    def __init__(self, file_path: str, parser: QueryParser,
                 collection_name: str='slots',
                 file_suffix: str='') -> None:
        super().__init__(file_path, parser, collection_name,
                         Slot, file_suffix)


class JsonTaskRepository(
        JsonRepository[Task], TaskRepository):
    """Json Task Repository"""

    def __init__(self, file_path: str, parser: QueryParser,
                 collection_name: str='tasks',
                 file_suffix: str='') -> None:
        super().__init__(file_path, parser, collection_name,
                         Task, file_suffix)



from ...application.domain.models import Classification
from ...application.domain.common import QueryParser
from ...application.domain.repositories import ClassificationRepository
from .rst_repository import RstRepository, RstLoader


class RstClassificationRepository(
        RstRepository[Classification], ClassificationRepository):
    """Restructuredtext Classification Repository"""

    def __init__(self, parser: QueryParser, loader: RstLoader) -> None:
        super().__init__(parser=parser, loader=loader,
                         item_class=Classification)
        self.counter = 0

    def load(self):
        for value in self.loader.nodes:
            classifier_ids = value.get('classifiers', [])
            if isinstance(classifier_ids, str):
                classifier_ids = [classifier_ids]

            for classifier_id in classifier_ids:
                self.counter += 1
                item = self.item_class(
                    **{'id': self.counter,
                        'classifier_id': classifier_id,
                       'task_id': value['id']})
                self.data[self._location][
                    str(self.counter)] = item  # type: ignore




# Base


class ApplicationError(Exception):
    """Application's base error class."""


# Repositories


class RepositoryError(ApplicationError):
    """Repositories' base error class."""


class EntityNotFoundError(RepositoryError):
    """The entity was not found in the repository."""


class EntityValidationError(RepositoryError):
    """Entity consistency validation error"""


# Coordinators


class CoordinatorError(ApplicationError):
    """Coordinators' base error class."""


class DataValidationError(CoordinatorError):
    """Data Validation Error."""



from .init_json_database import init_json_database
from .json_repository import JsonRepository
from .json_model_repositories import (
    JsonClassificationRepository, JsonClassifierRepository,
    JsonLinkRepository, JsonScheduleRepository,
    JsonSlotRepository, JsonTaskRepository)



from docutils.parsers.rst import Parser  # type: ignore
from docutils.frontend import OptionParser
from docutils.utils import new_document
from docutils.nodes import Node, NodeVisitor, field, title  # type: ignore


class RstAnalyzer:

    def analyze(self, content: str):
        parser = Parser()
        components = (Parser,)
        settings = OptionParser(components=components).get_default_values()
        document = new_document('<rst-doc>', settings=settings)
        parser.parse(content, document)

        result = {}

        def assign_result(value):
            entries = value.split('\n')
            for entry in entries:
                key, value = entry.split('=')
                value = value.strip()
                if ',' in value:
                    value = [item.strip() for item in value.split(',')]
                result[key.strip()] = value

        visitor = _RstVisitor(document, callback=assign_result)
        document.walk(visitor)
        return result


class _RstVisitor(NodeVisitor):

    def __init__(self, *attrs, tag: str = 'estimark', callback=None) -> None:
        super().__init__(*attrs)
        self.tag = tag
        self.callback = callback
        self.title: str = ''

    def visit_title(self, node) -> None:
        if not self.title and self.callback:
            self.title = "name={0}".format(node[0])
            self.callback(self.title)

    def visit_field(self, node) -> None:
        """Called for "reference" nodes."""
        key = node[0][0]
        if key == self.tag and self.callback:
            value = node[1][0][0]
            self.callback(value)

    def unknown_visit(self, node: Node) -> None:
        """Called for all other node types."""



from ...application.domain.models import Link
from ...application.domain.common import QueryParser
from ...application.domain.repositories import LinkRepository
from .rst_repository import RstRepository, RstLoader


class RstLinkRepository(RstRepository[Link], LinkRepository):
    """Restructuredtext Link Repository"""

    def __init__(self, parser: QueryParser,
                 loader: RstLoader) -> None:
        super().__init__(parser=parser, loader=loader, item_class=Link)
        self.counter = 0

    def load(self):
        nodes = self.loader.nodes

        previous = ''
        for value in nodes:
            if value.get('summary'):
                continue

            target = value.get('id')
            predecessors = value.get('predecessors', [])

            if not predecessors:
                predecessors.append(previous)
            elif not isinstance(predecessors, list):
                predecessors = [predecessors]

            for source in predecessors:
                self.counter += 1
                item = self.item_class(
                    id=self.counter, source=source, target=target)
                self.data[self._location][
                    str(self.counter)] = item # type: ignore

            previous = target



import os
from typing import Dict, Any


Config = Dict[str, Any]


config: Config = {
    "factory": os.environ.get('ESTIMARK_FACTORY', 'RstFactory'),
    "strategies": os.environ.get(
        'ESTIMARK_STRATEGIES', 'base,altair,json,rst').split(','),
    "root_dir": os.environ.get('ESTIMARK_ROOT_DIR', '.'),
    "param_dir": os.environ.get('ESTIMARK_PARAM_DIR', '.estimark'),
    "result_dir": os.environ.get('ESTIMARK_RESULT_DIR', '.estimark'),
    "plot_dir": os.environ.get('ESTIMARK_PLOT_DIR', '.estimark'),
}



from .entity import Entity


class Link(Entity):
    def __init__(self, **attributes):
        super().__init__(**attributes)
        self.source = attributes['source']
        self.target = attributes['target']



from typing import List, Union, Tuple, MutableMapping, Any

TermTuple = Tuple[str, str, Union[str, int, float, bool, list, tuple]]

QueryDomain = List[Union[str, TermTuple]]

DataDict = MutableMapping[str, Any]

RecordList = List[DataDict]



from ..core.common import Config
from ..application.domain.common import QueryParser
from ..application.domain.repositories import (
    TaskRepository, LinkRepository, ClassifierRepository,
    ClassificationRepository, ScheduleRepository, SlotRepository)
from ..core.data.json import (
    init_json_database, JsonRepository,
    JsonClassificationRepository, JsonClassifierRepository,
    JsonLinkRepository, JsonScheduleRepository,
    JsonSlotRepository, JsonTaskRepository)
from .altair_factory import AltairFactory


class JsonFactory(AltairFactory):
    def __init__(self, config: Config) -> None:
        super().__init__(config)
        self.config = config
        self.param_dir = self.config['param_dir']
        self.result_dir = self.config['result_dir']

    def classification_repository(
        self, query_parser: QueryParser,
    ) -> ClassificationRepository:
        return JsonClassificationRepository(
            self.param_dir, query_parser, file_suffix='param')

    def classifier_repository(
        self, query_parser: QueryParser,
    ) -> ClassifierRepository:
        return JsonClassifierRepository(
            self.param_dir, query_parser, file_suffix='param')

    def link_repository(
        self, query_parser: QueryParser,
    ) -> LinkRepository:
        return JsonLinkRepository(
            self.param_dir, query_parser, file_suffix='param')

    def schedule_repository(
        self, query_parser: QueryParser,
    ) -> ScheduleRepository:
        return JsonScheduleRepository(
            self.result_dir, query_parser, file_suffix='result')

    def slot_repository(
        self, query_parser: QueryParser,
    ) -> SlotRepository:
        return JsonSlotRepository(
            self.result_dir, query_parser, file_suffix='result')

    def task_repository(
        self, query_parser: QueryParser,
    ) -> TaskRepository:
        return JsonTaskRepository(
            self.param_dir, query_parser, file_suffix='param')



from .entity import Entity


class Schedule(Entity):
    def __init__(self, **attributes):
        super().__init__(**attributes)
        self.name = attributes.get('name')
        self.state = attributes.get('state', '')



from abc import ABC, abstractmethod
from ..domain.repositories import (
    TaskRepository, LinkRepository, ClassifierRepository,
    ScheduleRepository, SlotRepository)
from ..domain.common import QueryDomain, RecordList


class EstimarkInformer(ABC):

    @abstractmethod
    async def search(self,
                     model: str,
                     domain: QueryDomain = None,
                     limit: int = 0,
                     offset: int = 0) -> RecordList:
        """Returns a list of <<model>> dictionaries matching the domain"""


class StandardEstimarkInformer(EstimarkInformer):
    def __init__(self, task_repository: TaskRepository,
                 link_repository: LinkRepository,
                 classifier_repository: ClassifierRepository,
                 schedule_repository: ScheduleRepository,
                 slot_repository: SlotRepository) -> None:
        self.task_repository = task_repository
        self.link_repository = link_repository
        self.classifier_repository = classifier_repository
        self.schedule_repository = schedule_repository
        self.slot_repository = slot_repository

    def search(self,
               model: str,
               domain: QueryDomain = None,
               limit: int = 1000,
               offset: int = 0) -> RecordList:
        repository = getattr(self, f'{model}_repository')
        return [vars(entity) for entity in
                repository.search(
                    domain or [], limit=limit, offset=offset)]



"""
Estimark entrypoint
"""

import os
import sys
import logging
from json import loads
from pathlib import Path
from typing import List
from injectark import Injectark
from estimark.factories import factory_builder
from estimark.presenters.shell import Shell
from estimark.core.common import config


def main(args: List[str] = None):  # pragma: no cover
    args = args or sys.argv[1:]
    logging.basicConfig(stream=sys.stdout, level=logging.DEBUG,
                        format='%(message)s')

    factory = factory_builder.build(config)

    injector = Injectark(factory)

    Shell(config, injector).run(args)


if __name__ == '__main__':  # pragma: no cover
    sys.exit(main())



from .repository import Repository
from .memory_repository import MemoryRepository
from .memory_model_repositories import (
    ClassificationRepository, MemoryClassificationRepository,
    ClassifierRepository, MemoryClassifierRepository,
    LinkRepository, MemoryLinkRepository,
    ScheduleRepository, MemoryScheduleRepository,
    SlotRepository, MemorySlotRepository,
    TaskRepository, MemoryTaskRepository)



from typing import List, Union, Callable, Sequence
from fnmatch import fnmatchcase
from .types import TermTuple


class QueryParser:
    def __init__(self) -> None:
        self.comparison_dict = {
            '=': lambda field, value: (
                lambda obj: getattr(obj, field) == value),
            '!=': lambda field, value: (
                lambda obj: getattr(obj, field) != value),
            '<=': lambda field, value: (
                lambda obj: getattr(obj, field) <= value),
            '<': lambda field, value: (
                lambda obj: getattr(obj, field) < value),
            '>': lambda field, value: (
                lambda obj: getattr(obj, field) > value),
            '>=': lambda field, value: (
                lambda obj: getattr(obj, field) >= value),
            'in': lambda field, value: (
                lambda obj: getattr(obj, field) in value),
            'like': lambda field, value: (
                lambda obj: self._parse_like(getattr(obj, field), value)),
            'ilike': lambda field, value: (
                lambda obj: self._parse_like(
                    getattr(obj, field), value, True)),
            'contains': lambda field, value: (
                lambda obj: value in getattr(obj, field)),
        }

        self.binary_dict = {
            '&': lambda expression_1, expression_2: (
                lambda obj: (expression_1(obj) and expression_2(obj))),
            '|': lambda expression_1, expression_2: (
                lambda obj: (expression_1(obj) or expression_2(obj)))
        }

        self.unary_dict = {
            '!': lambda expression_1: (
                lambda obj: (not expression_1(obj)))
        }

        self.default_join_operator = '&'

    def parse(self, domain: Sequence[Union[str, TermTuple]]) -> Callable:
        if not domain:
            return lambda obj: True
        stack = []  # type: List[Callable]
        for item in list(reversed(domain)):
            if isinstance(item, str) and item in self.binary_dict:
                first_operand = stack.pop()
                second_operand = stack.pop()
                function = self.binary_dict[str(item)](
                    first_operand, second_operand)
                stack.append(function)
            elif isinstance(item, str) and item in self.unary_dict:
                operand = stack.pop()
                stack.append(self.unary_dict[str(item)](operand))

            stack = self._default_join(stack)

            if isinstance(item, (list, tuple)):
                result = self._parse_term(item)
                stack.append(result)

        result = self._default_join(stack)[0]
        return result

    def _default_join(self, stack: List[Callable]) -> List[Callable]:
        operator = self.default_join_operator
        if len(stack) == 2:
            first_operand = stack.pop()
            second_operand = stack.pop()
            function = self.binary_dict[operator](
                first_operand, second_operand)
            stack.append(function)
        return stack

    def _parse_term(self, term_tuple: TermTuple) -> Callable:
        field, operator, value = term_tuple
        function = self.comparison_dict.get(operator)
        result = function(field, value)
        return result

    @staticmethod
    def _parse_like(value: str, pattern: str, insensitive=False) -> bool:
        if not isinstance(value, str):
            return False
        pattern = pattern.replace('%', '*').replace('_', '?')
        if insensitive:
            pattern = pattern.lower()
            value = value.lower()
        return fnmatchcase(value, pattern)



from typing import List, Dict, Any
from ..domain.common import QueryDomain
from ..domain.models import Classifier
from ..domain.repositories import ClassifierRepository
from ..domain.services import PlotService


class InitializationManager:
    def __init__(self, classifier_repository: ClassifierRepository) -> None:
        self.classifier_repository = classifier_repository

    def initialize(self) -> None:
        classifiers = self.classifier_repository.search([])
        if classifiers:
            return

        self.classifier_repository.add([
            Classifier(id='XS', name='Extra Small', amount=1),
            Classifier(id='S', name='Small', amount=2),
            Classifier(id='M', name='Medium', amount=3),
            Classifier(id='L', name='Large', amount=5),
            Classifier(id='XL', name='Extra Large', amount=8),
            Classifier(id='DAY', name='Day', amount=1),
            Classifier(id='WEEK', name='Week', amount=7),
            Classifier(id='MONTH', name='Month', amount=30),
            Classifier(id='SEM', name='SEM', amount=180)
        ])



from collections import defaultdict
from typing import List, Dict, Type, Callable, Union, Generic
from ...application.domain.common import QueryParser, QueryDomain
from ...application.domain.models import T
from ...application.domain.repositories import Repository
from .rst_loader import RstLoader


class RstRepository(Repository, Generic[T]):
    def __init__(self, parser: QueryParser,
                 loader: RstLoader,
                 item_class: Type[T]) -> None:
        self.data: Dict[str, Dict[str, T]] = defaultdict(dict)
        self.parser = parser
        self.loader = loader
        self.item_class: Callable[..., T] = item_class
        self.max_items = 10_000

    def add(self, item: Union[T, List[T]]) -> List[T]:
        raise NotImplementedError('Implementation not available.')

    def search(self, domain: QueryDomain,
               limit=10_000, offset=0) -> List[T]:
        items = []
        filter_function = self.parser.parse(domain)
        for item in list(self.data[self._location].values()):
            if filter_function(item):
                items.append(item)

        if offset is not None:
            items = items[offset:]

        if limit is not None:
            items = items[:min(limit, self.max_items)]

        return items

    def remove(self, item: Union[T, List[T]]) -> bool:
        raise NotImplementedError('Implementation not available.')

    def count(self, domain: QueryDomain = None) -> int:
        count = 0
        domain = domain or []
        filter_function = self.parser.parse(domain)
        for item in list(self.data[self._location].values()):
            if filter_function(item):
                count += 1
        return count

    def load(self):
        for value in self.loader.nodes:
            item = self.item_class(**value)
            self.data[self._location][value['id']] = item

    @property
    def _location(self) -> str:
        return 'data'



from injectark import FactoryBuilder
from .base_factory import BaseFactory
from .altair_factory import AltairFactory
from .check_factory import CheckFactory
from .rst_factory import RstFactory
from .json_factory import JsonFactory


factory_builder = FactoryBuilder([
    BaseFactory, CheckFactory, AltairFactory, JsonFactory, RstFactory])



from .entity import Entity


class Classification(Entity):
    def __init__(self, **attributes):
        super().__init__(**attributes)
        self.classifier_id = attributes['classifier_id']
        self.task_id = attributes['task_id']



from .rst_repository import RstRepository
from .rst_analyzer import RstAnalyzer
from .rst_task_repository import RstTaskRepository
from .rst_loader import RstLoader
from .rst_link_repository import RstLinkRepository
from .rst_classification_repository import RstClassificationRepository



import logging
from json import loads
from argparse import ArgumentParser, Namespace
from typing import Dict, List
from injectark import Injectark
from ... import __version__
from ...core import Config


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger()


class Shell:
    def __init__(self, config: Config, injector: Injectark) -> None:
        self.injector = injector
        self.config = config

    def run(self, args):
        namespace = self.parse(args)
        namespace.func(vars(namespace))

    def parse(self, args) -> Namespace:
        parser = ArgumentParser('Estimark')
        subparsers = parser.add_subparsers(dest='action')

        # Initialize
        estimate_parser = subparsers.add_parser('init')
        estimate_parser.set_defaults(func=self.init)

        # Estimate
        estimate_parser = subparsers.add_parser('estimate')
        estimate_parser.add_argument('-s', '--states', default='')
        estimate_parser.set_defaults(func=self.estimate)

        # Show
        show_parser = subparsers.add_parser('show')
        show_parser.add_argument('-m', '--model', default='task')
        show_parser.set_defaults(func=self.show)

        # Plot
        plot_parser = subparsers.add_parser('plot')
        plot_parser.add_argument('-t', '--type', default='gantt')
        plot_parser.add_argument('-c', '--context', default='{}',
                                 help='JSON plot parameters context.')
        plot_parser.set_defaults(func=self.plot)

        # Version
        version_parser = subparsers.add_parser('version')
        version_parser.set_defaults(func=self.version)

        if len(args) == 0:
            parser.print_help()
            parser.exit()

        return parser.parse_args(args)

    def init(self, options_dict: Dict[str, str]) -> None:
        logger.info('<< INIT >>')
        initialization_manager = self.injector['InitializationManager']
        initialization_manager.initialize()

    def estimate(self, options_dict: Dict[str, str]) -> None:
        logger.info('<< ESTIMATE >>')
        states = [state.strip() for state in
                  options_dict.get('states', '').split(',') if state]
        estimation_manager = self.injector['EstimationManager']
        estimation_manager.estimate(states)

    def show(self, options_dict: Dict[str, str]) -> None:
        logger.info('<< SHOW >>')
        estimark_informer = self.injector['EstimarkInformer']
        models = ['task', 'link', 'classifier', 'schedule', 'slot']
        model = options_dict.get('model')
        if model not in models:
            raise ValueError(
                f"The model should be one of: {', '.join(models)}")

        records = estimark_informer.search(model, [])

        logger.info(f'{model} records:')
        for record in records:
            logger.info(record)

    def plot(self, options_dict: Dict[str, str]):
        logger.info('<< PLOT >>')
        estimation_manager = self.injector['EstimationManager']
        type_ = options_dict['type']
        context = loads(options_dict['context'])
        estimation_manager.plot(type_, context)

    def version(self, options_dict: Dict[str, str]):
        logger.info('<< VERSION >>')
        logger.info(__version__)



from ...application.domain.models import Task
from ...application.domain.common import QueryParser
from ...application.domain.repositories import TaskRepository
from .rst_repository import RstRepository, RstLoader


class RstTaskRepository(RstRepository[Task], TaskRepository):
    """Restructuredtext Task Repository"""

    def __init__(self, parser: QueryParser,
                 loader: RstLoader) -> None:
        super().__init__(parser=parser, loader=loader, item_class=Task)



from .entity import Entity


class Classifier(Entity):
    def __init__(self, **attributes):
        super().__init__(**attributes)
        self.name = attributes['name']
        self.amount = attributes['amount']
        self.units = attributes.get('units', 'days')



from ..core.common import Config
from ..application.domain.models import Task
from ..application.domain.common import QueryParser
from ..application.domain.repositories import (
    TaskRepository, MemoryTaskRepository)
from .base_factory import BaseFactory


class CheckFactory(BaseFactory):
    def __init__(self, config: Config) -> None:
        super().__init__(config)
        self.config = config

    # Repositories
    def task_repository(
            self, query_parser: QueryParser
    ) -> TaskRepository:

        task_repository = super().task_repository(query_parser)

        task_repository.load({'data': {
            "1": Task(id='1', name='Define WBS'),
            "2": Task(id='2', name='Deploy Servers'),
            "3": Task(id='3', name='Design Website')
        }})
        return task_repository



import json
import logging
from typing import Dict, List
from abc import ABC, abstractmethod
from ..models import Schedule, Slot, Task
from ..repositories import SlotRepository


logger = logging.getLogger(__name__)


class PlotService(ABC):
    @abstractmethod
    def plot_gantt(self, schedule: Schedule) -> str:
        "Plot gantt method to be implemented."

    @abstractmethod
    def plot_kanban(self, tasks: List[Task], sort: List[str] = None,
                    group: str = None) -> str:
        "Plot kanban method to be implemented."


class MemoryPlotService(PlotService):
    def __init__(self) -> None:
        self.gantt_plotted = False
        self.kanban_plotted = False

    def plot_gantt(self, schedule: Schedule) -> None:
        self.gantt_plotted = True
        logging.info(f"MEMORY PLOT. SCHEDULE: {schedule.name}")

    def plot_kanban(self, tasks: List[Task], sort: List[str] = None,
                    group: str = None) -> None:
        self.kanban_plotted = True
        logging.info(f"MEMORY KANBAN PLOT. TASKS #: {len(tasks)}")



from injectark import Factory
from ..core.common import Config
from ..application.domain.common import QueryParser
from ..application.domain.repositories import (
    TaskRepository, MemoryTaskRepository,
    LinkRepository, MemoryLinkRepository,
    ClassifierRepository, MemoryClassifierRepository,
    ClassificationRepository, MemoryClassificationRepository,
    ScheduleRepository, MemoryScheduleRepository,
    SlotRepository, MemorySlotRepository)
from ..application.domain.services import PlotService, MemoryPlotService
from ..application.managers import EstimationManager, InitializationManager
from ..application.informers import EstimarkInformer, StandardEstimarkInformer


class BaseFactory(Factory):
    def __init__(self, config: Config) -> None:
        super().__init__(config)
        self.config = config

    def query_parser(self) -> QueryParser:
        return QueryParser()

    def task_repository(
            self, query_parser: QueryParser
    ) -> TaskRepository:
        return MemoryTaskRepository(query_parser)

    def link_repository(
            self, query_parser: QueryParser
    ) -> LinkRepository:
        return MemoryLinkRepository(query_parser)

    def classifier_repository(
            self, query_parser: QueryParser
    ) -> ClassifierRepository:
        return MemoryClassifierRepository(query_parser)

    def classification_repository(
            self, query_parser: QueryParser
    ) -> ClassificationRepository:
        return MemoryClassificationRepository(query_parser)

    def schedule_repository(
            self, query_parser: QueryParser
    ) -> ScheduleRepository:
        return MemoryScheduleRepository(query_parser)

    def slot_repository(
            self, query_parser: QueryParser
    ) -> SlotRepository:
        return MemorySlotRepository(query_parser)

    def plot_service(self) -> PlotService:
        return MemoryPlotService()

    def estimation_manager(
            self, task_repository: TaskRepository,
            classifier_repository: ClassifierRepository,
            classification_repository: ClassificationRepository,
            link_repository: LinkRepository,
            schedule_repository: ScheduleRepository,
            slot_repository: SlotRepository,
            plot_service: PlotService
    ) -> EstimationManager:
        return EstimationManager(
            task_repository, classifier_repository,
            classification_repository, link_repository,
            schedule_repository, slot_repository, plot_service)

    def initialization_manager(
            self, classifier_repository: ClassifierRepository,
    ) -> InitializationManager:
        return InitializationManager(classifier_repository)

    def estimark_informer(
            self, task_repository: TaskRepository,
            link_repository: LinkRepository,
            classifier_repository: ClassifierRepository,
            schedule_repository: ScheduleRepository,
            slot_repository: SlotRepository
    ) -> EstimarkInformer:
        return StandardEstimarkInformer(
            task_repository, link_repository, classifier_repository,
            schedule_repository, slot_repository)



from .entity import Entity


class Slot(Entity):
    def __init__(self, **attributes):
        super().__init__(**attributes)
        self.name = attributes.get('name', '')
        self.task_id = attributes['task_id']
        self.schedule_id = attributes['schedule_id']
        self.start = attributes['start']
        self.end = attributes['end']



from typing import List, Dict, Any
from pathlib import Path
from .rst_analyzer import RstAnalyzer


class RstLoader:

    def __init__(self, root: str, analyzer: RstAnalyzer) -> None:
        self._nodes: List[Dict[str, Any]] = []
        self.root = root
        self.analyzer = analyzer

    @property
    def nodes(self):
        if not self._nodes:
            self.load()
        return self._nodes

    def load(self):
        path = Path(self.root)

        nodes_dict = {}
        for node in sorted(path.rglob('*.rst')):
            data = self._analyze_node(node)

            if 'index.rst' in node.name:
                node = node.parent
                data['summary'] = True

            metadata = {
                'file_name': node.name,
                'parent_dir': node.parent.name,
                'parent_absolute': str(node.parent.absolute())
            }

            nodes_dict[str(node.absolute())] = {**metadata, **data}

        for value in nodes_dict.values():
            value['id'] = self._extract_id(value)
            value['parent_id'] = self._extract_parent_id(nodes_dict, value)
            self._nodes.append(value)

    def _analyze_node(self, node: Path) -> Dict[str, Any]:
        data_dict = self._extract_content(node)
        return data_dict

    def _extract_content(self, node: Path) -> Dict[str, Any]:
        content = node.read_text()
        return self.analyzer.analyze(content)

    def _extract_id(self, value: Dict[str, Any]) -> str:
        _id = value.get('id', '')
        if _id:
            return _id

        file_name = value.get('file_name', '')
        parent_dir = value.get('parent_dir', '')

        if '_' in file_name:
            _id, *rest = file_name.split('_')
        elif '_' in parent_dir:
            _id, *rest = parent_dir.split('_')

        return _id

    def _extract_parent_id(self, nodes_dict: Dict[str, Any],
                           value: Dict[str, Any]) -> str:
        parent_value = nodes_dict.get(value.get('parent_absolute', ''))
        if not parent_value:
            return ''

        return self._extract_id(parent_value)




import altair as alt
import pandas as pd
import logging
from typing import List
from datetime import date
from pathlib import Path
from collections import Counter
from ...application.domain.models import Schedule, Task
from ...application.domain.repositories import SlotRepository
from ...application.domain.services import PlotService


logger = logging.getLogger(__name__)


class AltairPlotService(PlotService):
    def __init__(self, plot_dir: str,
                 slot_repository: SlotRepository) -> None:
        self.slot_repository = slot_repository
        self.plot_dir = plot_dir

    def plot_gantt(self, schedule: Schedule) -> None:
        logging.info(f"ALTAIR PLOT SCHEDULE: {schedule.name}")
        slots = self.slot_repository.search(
            [('schedule_id', '=', schedule.id)])
        slot_dict_list = [vars(slot) for slot in slots]
        state = f"| State: <{schedule.state}> " if schedule.state else ""
        title = f"{schedule.name} {state}| {date.today()}"

        source = pd.DataFrame(slot_dict_list)
        chart = alt.Chart(source).mark_bar().encode(
            x='start',
            x2='end',
            y='name'
        ).properties(title=title)

        output_file = str(Path(self.plot_dir).joinpath('gantt.html'))

        chart.save(output_file)

    def plot_kanban(self, tasks: List[Task],
                    sort: List[str] = None, group: str = None) -> None:
        logging.info(f"ALTAIR KANBAN PLOT. TASKS #: {len(tasks)}")
        group = group or 'state'
        detail = 'owner' if group == 'state' else 'state'

        task_dict_list = [{'weight': 1, **vars(task)}
                          for task in tasks if not task.summary]

        counts = Counter(task_dict[group] for task_dict
                         in task_dict_list).most_common()
        _, max_depth = next(iter(counts))

        source = pd.DataFrame(task_dict_list)

        title = f"Kanban Chart | {date.today()}"
        block_height = 50
        block_width = block_height * 8

        base = alt.Chart(source).mark_bar(
            color='black'
        ).encode(
            x=alt.X(group, axis=alt.Axis(
                orient='top', labelAngle=0, labelFontSize=15), sort=sort),
            y=alt.Y('sum(weight)', sort='descending', stack='zero'),
            order=alt.Order('id',  sort='ascending')
        ).properties(
            title=title,
            width=block_width * len(counts),
            height=block_height * max_depth)

        bars = base.encode(
            color=alt.Color('id:N', legend=None))

        text = base.mark_text(
            dy=-(block_height * 0.33),
            color='black'
        ).encode(
            text='name')

        info = base.mark_text(
            dy=-(block_height * 0.67),
            dx=(block_width * 0.3),
            color='#2F4F4F'
        ).encode(
            text=detail)

        chart = bars + text + info

        output_file = str(Path(self.plot_dir).joinpath('kanban.html'))

        chart.save(output_file)



from typing import List, Dict, Any
from ..domain.common import QueryDomain
from ..domain.models import Schedule, Slot
from ..domain.repositories import (
    TaskRepository, ClassifierRepository, ClassificationRepository,
    LinkRepository, SlotRepository, ScheduleRepository)
from ..domain.services import PlotService


class EstimationManager:
    def __init__(self, task_repository: TaskRepository,
                 classifier_repository: ClassifierRepository,
                 classification_repository: ClassificationRepository,
                 link_repository: LinkRepository,
                 schedule_repository: ScheduleRepository,
                 slot_repository: SlotRepository,
                 plot_service: PlotService
                 ) -> None:
        self.task_repository = task_repository
        self.classifier_repository = classifier_repository
        self.classification_repository = classification_repository
        self.link_repository = link_repository
        self.schedule_repository = schedule_repository
        self.slot_repository = slot_repository
        self.plot_service = plot_service

    def estimate(self, states: List[str] = None):
        slot_dict_list = self._calculate_slots(states)

        schedule, *_ = self.schedule_repository.add(
            Schedule(name='Project Schedule', state=", ".join(states or [])))

        for slot_dict in slot_dict_list:
            slot_dict.update({'schedule_id': schedule.id})
            self.slot_repository.add(Slot(**slot_dict))

    def plot(self, type: str = 'gantt',
             context: Dict[str, Any] = None) -> bool:
        context = context or {}
        plot_types = ['gantt', 'kanban']
        if type not in plot_types:
            raise ValueError(f'The plot type should be one of: {plot_types}')

        if type == 'kanban':
            domain: QueryDomain = []
            sort = None
            if context.get('states'):
                sort = context['states']
                domain.append(('state', 'in', sort))
            if context.get('owners'):
                sort = context['owners']
                domain.append(('owner', 'in', sort))
            tasks = self.task_repository.search(domain)
            if not tasks:
                return False
            self.plot_service.plot_kanban(
                tasks, sort, context.get('group'))
        else:
            schedule_id = context.get('schedule_id')
            domain = [('id', '=', schedule_id)] if schedule_id else []
            schedule = next(iter(
                self.schedule_repository.search(domain)), None)
            if not schedule:
                return False
            self.plot_service.plot_gantt(schedule)

        return True

    def _calculate_slots(self, states: List[str] = None
                         ) -> List[Dict[str, Any]]:
        domain: QueryDomain = [('summary', '=', False)]
        if states:
            domain.append(('state', 'in', states))
        effective_tasks = self.task_repository.search(domain)

        slots = []
        task_amounts_dict = {}
        for task in effective_tasks:
            task_amounts_dict[task.id] = {
                'name': task.name,
                'amount':  self._calculate_amount(task.id)
            }

        for task_id, values in task_amounts_dict.items():
            start = self._calculate_start(task_amounts_dict, task_id)
            end = start + values['amount']
            slots.append({'name': values['name'], 'task_id': task_id,
                          'start': start, 'end': end})

        return slots

    def _calculate_start(self, task_amounts_dict, task_id):
        durations = [0]
        predecessor_ids = [
            predecessor.source for predecessor in
            self.link_repository.search([('target', '=', task_id)])]
        for predecessor_id in predecessor_ids:
            duration = self._calculate_start(task_amounts_dict, predecessor_id)
            duration += task_amounts_dict.get(
                predecessor_id, {}).get('amount', 0)
            durations.append(duration)
        return max(durations)

    def _calculate_amount(self, task_id: str) -> float:
        classifier_ids = [
            classification.classifier_id for classification in
            self.classification_repository.search(
                [('task_id', '=', task_id)])
        ]
        classifiers = self.classifier_repository.search(
            [('id', 'in', classifier_ids)])
        amount = sum([classifier.amount for classifier in classifiers])
        return amount



import os
from json import dump
from typing import Dict, Any


DEFAULT_SCHEMA = {
    "schedules": {},
    "slots": {}
}


def init_json_database(
        path: str, schema: Dict[str, Any] = DEFAULT_SCHEMA) -> bool:
    if os.path.exists(path):
        return False
    with open(path, 'w') as f:
        dump(schema, f)  # type: ignore
    return True



import os
import time
from uuid import uuid4
from pathlib import Path
from json import loads, load, dump
from collections import defaultdict
from typing import Dict, List, Union, Any, Type, TypeVar, Callable, Generic
from ....application.domain.models import T
from ....application.domain.common import QueryParser, QueryDomain
from ....application.domain.repositories import Repository


class JsonRepository(Repository, Generic[T]):
    def __init__(self, directory_path: str, parser: QueryParser,
                 collection_name: str, item_class: Type[T],
                 file_suffix: str = '') -> None:
        self.directory_path = directory_path
        self.parser = parser
        self.collection_name = collection_name
        self.item_class: Callable[..., T] = item_class
        self.file_suffix = file_suffix
        self.max_items = 10_000

    def add(self, item: Union[T, List[T]]) -> List[T]:
        items = item if isinstance(item, list) else [item]
        data: Dict[str, Any] = defaultdict(lambda: {})
        if self.file_path.exists():
            data.update(loads(self.file_path.read_text()))

        for item in items:
            item.id = item.id or str(uuid4())
            item.updated_at = int(time.time())
            if not data[self.collection_name].get(item.id):
                item.created_at = item.updated_at

            data[self.collection_name][item.id] = vars(item)

        self.file_path.parent.mkdir(parents=True, exist_ok=True)
        with self.file_path.open('w') as f:
            dump(data, f, indent=2)

        return items

    def search(self, domain: QueryDomain,
               limit=10_000, offset=0) -> List[T]:
        if not self.file_path.exists():
            return []

        with self.file_path.open('r') as f:
            data = load(f)
            items_dict = data.get(self.collection_name, {})

        items = []
        filter_function = self.parser.parse(domain)
        for item_dict in items_dict.values():
            item = self.item_class(**item_dict)
            if filter_function(item):
                items.append(item)

        if offset is not None:
            items = items[offset:]

        if limit is not None:
            items = items[:min(limit, self.max_items)]

        return items

    def remove(self, item: Union[T, List[T]]) -> bool:
        items = item if isinstance(item, list) else [item]
        if not self.file_path.exists():
            return False

        with self.file_path.open('r') as f:
            data = load(f)

        deleted = False
        for item in items:
            deleted_item = data[self.collection_name].pop(item.id, None)
            deleted = bool(deleted_item) or deleted

        with self.file_path.open('w') as f:
            dump(data, f)

        return deleted

    def count(self, domain: QueryDomain = None) -> int:
        if not self.file_path.exists():
            return 0

        with self.file_path.open('r') as f:
            data = load(f)

        count = 0
        domain = domain or []
        filter_function = self.parser.parse(domain)
        for item_dict in list(data[self.collection_name].values()):
            item = self.item_class(**item_dict)
            if filter_function(item):
                count += 1
        return count

    @property
    def _location(self) -> str:
        return 'data'

    @property
    def file_path(self) -> Path:
        file_path = f'{self.directory_path}/{self._location}'
        if self.file_suffix:
            file_path = f'{file_path}_{self.file_suffix}'
        return Path(f'{file_path}.json')



from ..core.common import Config
from ..application.domain.common import QueryParser
from ..application.domain.repositories import (
    TaskRepository, LinkRepository, ClassificationRepository)
from ..core.data import (
    RstTaskRepository, RstLinkRepository,
    RstAnalyzer, RstLoader, RstClassificationRepository)
from .json_factory import JsonFactory


class RstFactory(JsonFactory):
    def __init__(self, config: Config) -> None:
        super().__init__(config)
        self.config = config

    def rst_analyzer(self):
        return RstAnalyzer()

    def rst_loader(self, analyzer: RstAnalyzer):
        root = self.config['root_dir']
        return RstLoader(root, analyzer)

    def task_repository(
        self, query_parser: QueryParser, loader: RstLoader
    ) -> TaskRepository:
        repository = RstTaskRepository(query_parser, loader)
        repository.load()
        return repository

    def link_repository(
        self, query_parser: QueryParser, loader: RstLoader
    ) -> RstLinkRepository:
        repository = RstLinkRepository(query_parser, loader)
        repository.load()
        return repository

    def classification_repository(
        self, query_parser: QueryParser, loader: RstLoader
    ) -> ClassificationRepository:
        repository = RstClassificationRepository(query_parser, loader)
        repository.load()
        return repository



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['estimark',
 'estimark.application',
 'estimark.application.domain',
 'estimark.application.domain.common',
 'estimark.application.domain.models',
 'estimark.application.domain.repositories',
 'estimark.application.domain.services',
 'estimark.application.informers',
 'estimark.application.managers',
 'estimark.core',
 'estimark.core.common',
 'estimark.core.data',
 'estimark.core.data.json',
 'estimark.core.plot',
 'estimark.factories',
 'estimark.presenters',
 'estimark.presenters.shell']

package_data = \
{'': ['*']}

install_requires = \
['altair>=4.1.0,<5.0.0', 'docutils>=0.16,<0.17', 'injectark>=0.6.0,<0.7.0']

entry_points = \
{'console_scripts': ['estimark = estimark.__main__:main']}

setup_kwargs = {
    'name': 'estimark',
    'version': '0.5.1',
    'description': 'Frictionless Estimation',
    'long_description': None,
    'author': 'Esteban Echeverry',
    'author_email': 'eecheverry@knowark.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from .entity import Entity


class Task(Entity):
    def __init__(self, **attributes):
        super().__init__(**attributes)
        self.name = attributes.get('name', '')
        self.summary = attributes.get('summary', False)
        self.parent_id = attributes.get('parent_id')
        self.state = attributes.get('state', 'backlog')
        self.owner = attributes.get('owner', 'unassigned')



from abc import ABC, abstractmethod
from typing import List, Generic, Union, Any
from ..models import T
from ..common import QueryDomain


class Repository(ABC, Generic[T]):
    @abstractmethod
    def add(self, item: Union[T, List[T]]) -> List[T]:
        "Add method to be implemented."

    @abstractmethod
    def search(self, domain: QueryDomain,
               limit: int = None, offset: int = None) -> List[T]:
        "Search items matching a query domain"

    @abstractmethod
    def remove(self, item: Union[T, List[T]]) -> bool:
        "Remove method to be implemented."

    @abstractmethod
    def count(self, domain: QueryDomain = None) -> int:
        "Count items matching a query domain"



from typing import TypeVar


class Entity:
    def __init__(self, **attributes) -> None:
        self.id = attributes.get('id', '')
        self.created_at = attributes.get('created_at', 0)
        self.updated_at = attributes.get('updated_at', self.created_at)


T = TypeVar('T', bound=Entity, covariant=True)



from ..core.common import Config
from ..application.domain.services import PlotService
from ..application.domain.repositories import SlotRepository
from ..core.plot import AltairPlotService
from .base_factory import BaseFactory


class AltairFactory(BaseFactory):
    def __init__(self, config: Config) -> None:
        super().__init__(config)
        self.config = config
        self.plot_dir = self.config['plot_dir']

    def plot_service(
        self, slot_repository: SlotRepository
    ) -> PlotService:
        plot_service = AltairPlotService(
            self.plot_dir, slot_repository)
        return plot_service



from ..models import (
    Classification, Classifier, Link, Schedule, Slot, Task)
from .repository import Repository
from .memory_repository import MemoryRepository


class ClassificationRepository(Repository[Classification]):
    """Classification Repository"""


class MemoryClassificationRepository(
        MemoryRepository[Classification], ClassificationRepository):
    """Memory Classification Repository"""


class ClassifierRepository(Repository[Classifier]):
    """Classifier Repository"""


class MemoryClassifierRepository(
        MemoryRepository[Classifier], ClassifierRepository):
    """Memory Classifier Repository"""


class LinkRepository(Repository[Link]):
    """Link Repository"""


class MemoryLinkRepository(
        MemoryRepository[Link], LinkRepository):
    """Memory Link Repository"""


class ScheduleRepository(Repository[Schedule]):
    """Schedule Repository"""


class MemoryScheduleRepository(
        MemoryRepository[Schedule], ScheduleRepository):
    """Memory Schedule Repository"""


class SlotRepository(Repository[Slot]):
    """Slot Repository"""


class MemorySlotRepository(
        MemoryRepository[Slot], SlotRepository):
    """Memory Slot Repository"""


class TaskRepository(Repository[Task]):
    """Task Repository"""


class MemoryTaskRepository(
        MemoryRepository[Task], TaskRepository):
    """Memory Task Repository"""



from .entity import Entity, T
from .classifier import Classifier
from .task import Task
from .classification import Classification
from .schedule import Schedule
from .slot import Slot
from .link import Link
